//**************************************************************************/
// Copyright (c) 1998-2007 Autodesk, Inc.
// All rights reserved.
// 
// These coded instructions, statements, and computer programs contain
// unpublished proprietary information written by Autodesk, Inc., and are
// protected by Federal copyright law. They may not be disclosed to third
// parties or copied or duplicated in any form, in whole or in part, without
// the prior written consent of Autodesk, Inc.
//**************************************************************************/
// DESCRIPTION: Includes for Plugins
// AUTHOR: Felix Joleanes
//***************************************************************************/
#include "AnimatablePivot.h"
#include "ControlContainer.h"
#include "AnimPivotTarget.h"

#define ANIMPIVOT_INTERFACE Interface_ID(0x397011f5, 0x4dba371f)

#define FKCtrl_CLASS_ID			Class_ID(0xf266d92, 0x3c2f6126)
#define IKCtrl_CLASS_ID			Class_ID(0x521f77e0, 0x4f440f13)
#define LOCALCTRL_REF		1
 
// mixin interface 
class AnimPivotInterface : public FPMixinInterface 
{ 
public:
	enum { fnGetRelativePos, fnSetRelativePos, }; 
   BEGIN_FUNCTION_MAP 
     FN_0(fnGetRelativePos, TYPE_POINT3_BV, GetRelativePos); 
     VFN_1(fnSetRelativePos, SetRelativePos, TYPE_POINT3); 
   END_FUNCTION_MAP 
   FPInterfaceDesc* GetDesc();
   virtual Point3 GetRelativePos()=0; 
   virtual void SetRelativePos(Point3 relPos)=0; 
};


class AnimPivot : public Control, public AnimPivotInterface {
	public:
		void DeleteThis() { delete this; }	

		//references	
		ControlContainer *pivotCtrl;
		ControlContainer *TMCtrl;	
		IParamBlock2	*pblock;
		IParamBlock2	*stepPblock;

		//offset controllers
		ControlContainer *offsetCtrl;
		Control *prevPivotCtrl;
		Point3 SKMOffset;		//Set Key Mode Offset
		Point3 SKMPrevPivot;	//Set Key Mode prev Pivot 

		//Keyframe interfaces
		IKeyControl *pivotKeys;
		IKeyControl *prevPivotKeys;
		IKeyControl *offsetKeys;

		//AnimPivot variables
		Interval ivalid;
		TimeValue prevT;
		Matrix3 value;
		BOOL needUpdate;
		BOOL notIgnoreChange;
		BOOL pivotHasChanged;

		BOOL targetSetup;

		Matrix3 parentTM;

		//the next variable is used to dont allow offset update within a mouse cycle
		//if this value is 0 or 1 the offset update could be done, otherwise not.
		//an int is used instead of a bool to allow an offset update in the first change iteration
		//of the mouse cycle, this is done in this way in order to be able to make the
		//stepOffset of the new pivot keyframes (if a new keyframe is created within the mouse cycle)
		//unfortunatelly trackview interaction doesnt send mouse_cycle_started messages, so the offset
		//will be updated realtime while interacting, slowing down the system if there is toooo many keyframes
		//it also happen while slidding the spinners of transformation at the botton of the 3ds max interface
		//and the spinners of the transform popup dialog box (the one that appears when you right-click the move/rotate/scale buttons)
		int canUpdateOffset; 

		//step Offset variables
		//Tab<AnimPivotStepInterval*> stepIntervals;
		//Tab<BOOL> SOisKeyStep;

		IObjParam *ip;
		HWND hwnd;

		//Animatable Methods
		Class_ID ClassID() {return AnimPivot_CLASS_ID;}		
		SClass_ID SuperClassID() { return CTRL_MATRIX3_CLASS_ID; }
		void GetClassName(TSTR& s) {s = GetString(IDS_CLASS_NAME);}
		int NumSubs() { return NSUBS;}
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);
		int SubNumToRefNum(int subNum) { return (subNum);}
		void MouseCycleCompleted(TimeValue t);
		void MouseCycleStarted(TimeValue t);


		// Set-key mode support
		void RevertSetKeyBuffer();
		BOOL SetKeyBufferPresent() {return TRUE;}

		//to display controller params on the motion panel 
		void BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev ); 
		// to undisplay 
		void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next ); 
		// return number of ParamBlocks in this instance 
		int NumParamBlocks() {return NPBLOCKS;} // return i'th ParamBlock 
		IParamBlock2* GetParamBlock(int i) {if (i==0) return pblock; else return NULL;} 
		// return id'd ParamBlock 
		IParamBlock2* GetParamBlockByID(BlockID id) {
			if (pblock->ID()==id)
				return pblock;
			return NULL;
		}

		//Reference Methods
		int NumRefs() {return NREFS;}
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);

#if IS_SUPMAX2015
		RefResult NotifyRefChanged(const Interval& changeInt, RefTargetHandle hTarget, 
			PartID& partID,  RefMessage message, BOOL propagate);
#else
		RefResult NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
			PartID& partID,  RefMessage message);
#endif

		IOResult Load(ILoad *iload) {return IO_OK;}
		IOResult Save(ISave *isave) {return IO_OK;}
		void RescaleWorldUnits(float f) {}


		//Control Methods
		void Copy(Control *from);
		RefTargetHandle Clone( RemapDir &remap );
		int IsKeyable(){return 1;}
		BOOL IsLeaf(){return FALSE;}
		void GetValue(TimeValue t, void *val, Interval &valid, GetSetMethod method = CTRL_ABSOLUTE );
		void SetValue(TimeValue t, void *val, int commit=1, GetSetMethod method= CTRL_ABSOLUTE);
		void CommitValue(TimeValue t);
		void RestoreValue(TimeValue t);
		Control* GetPositionController();
		Control* GetRotationController();
		Control* GetScaleController();
		BOOL SetPositionController(Control *c);
		BOOL SetRotationController(Control *c);
		BOOL SetScaleController(Control *c);
		BOOL InheritsParentTransform (){return FALSE;}
		BOOL ChangeParents  (TimeValue  t,const Matrix3 &oldP, const Matrix3 &newP, const Matrix3 &tm);
		void GetWorldBoundBox  (  TimeValue  t,INode *  inode,ViewExp *  vpt,Box3 &  box)
		{
			parentTM=inode->GetParentTM(t);
		}

		//FPInterface methods
		FPInterface* GetInterface(Interface_ID id) { 
		 if (id == ANIMPIVOT_INTERFACE) 
		   return (AnimPivotInterface*)this; 
		 else 
			 return (FPInterface*)Control::GetInterface(id); 
		 } 

		//AnimPivot methods
		void CheckTarget();
		ReferenceTarget* FindAnimPivotTarget(ReferenceTarget* theTarget);
		void Update(TimeValue t);
		void InitOffset(Point3 offset);
		void UpdateOffset();
		void MakeOffset();
		void MakeOffset(int keyIndex);
		BOOL CopyOffset(int keyIndex);
		BOOL SwapOffset(int keyIndex);
		Point3 MakeOffset(Matrix3 TMVal,Point3 pivotVal,Point3 prevPivotVal,Point3 offsetVal);
		BOOL KeysEqual(int pivotKeyIndex,int prevPivotKeyIndex);
		void SetStepKey(int keyIndex);
		void ResetTM();
		void RemoveTarget();
		Matrix3 GetTargetTM(TimeValue t,Interval &valid);
		Matrix3 GetTargetTM(INode *theNode,TimeValue t,Interval &valid);

		//StepOffset Methods
		//these methods are used to establish an extra position for the pivot in step intervals (in step keyframes)
		//to avoid gaps in the node position when there is a step keyframe with a current TMCtrl value
		//diferent from the previous keyframe.
		//if a new keyframe is created in the pivotCtrl it's value needs to be changed to match this extra interpolation
		void SOOffsetKey(int theKeyIndex);
		Point3 SOGetStepBezier(float t,Point3 val);

		//published methods
		Point3 GetRelativePos(); 
		void SetRelativePos(Point3 relPos); 
		
		//Constructor/Destructor
		AnimPivot(BOOL loading=FALSE);
		~AnimPivot() {}
};

class AnimPivotClassDesc : public ClassDesc2 
{
public:
	virtual int IsPublic() 							{ return TRUE; }
	virtual void* Create(BOOL loading) 		{return new AnimPivot(loading); }
	void DeleteThis() { delete this; }
	virtual const TCHAR *	ClassName() 			{ return GetString(IDS_CLASS_NAME); }
	virtual SClass_ID SuperClassID() 				{ return CTRL_MATRIX3_CLASS_ID; }
	virtual Class_ID ClassID() 						{ return AnimPivot_CLASS_ID; }
	virtual const TCHAR* Category() 				{ return _T(""); }

	virtual const TCHAR* InternalName() 			{ return _T("AnimPivot"); }	// returns fixed parsable name (scripter-visible name)
	virtual HINSTANCE HInstance() 					{ return hInstance; }					// returns owning module handle
};

class AnimPivotDlgProc: public ParamMap2UserDlgProc
{
public:
	AnimPivot *animpivot;
	AnimPivotDlgProc(AnimPivot *animPivot){animpivot=animPivot;}
	virtual INT_PTR  DlgProc (TimeValue t, IParamMap2 *map, HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
	virtual void  DeleteThis (){delete this;}
};




class AnimPivotClassAccessor : public PBAccessor
{ 
	public:
		void Set(PB2Value& v,ReferenceMaker* owner,ParamID id,int tabIndex,TimeValue  t);
};
class AnimPivotClassValidator : public PBValidator
{ 
	public:
		BOOL Validate(PB2Value & v){return FALSE;}
		BOOL Validate(PB2Value& v,ReferenceMaker* owner,ParamID id,int tabIndex);
};

class AnimPivotPostPatchProc : public PostPatchProc
{
public:
	AnimPivot *ctrl;
	AnimPivotPostPatchProc(AnimPivot *theCtrl){ctrl=theCtrl;}
	int  Proc (RemapDir &remap);
};

